;; Generated package description from helm.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "helm" "4.0.7.0.20260728.35" "Helm is an Emacs incremental and narrowing framework" '((helm-core "4.0.7") (wfnames "1.2")) :commit "c438a08d1bbc2286e13891e4de812ef1b5190628" :authors '(("Thierry Volpiatto" . "thievol@posteo.net")) :maintainer '("Thierry Volpiatto" . "thievol@posteo.net") :keywords '("helm" "convenience" "files" "buffers" "grep" "completion" "lisp" "matching" "tools" "help" "package") :url "https://emacs-helm.github.io/helm/")
